/**
 * Copyright (C) 2004 Billy Biggs <vektor@dumbterm.net>.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdio.h>
#include <stdint.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <X11/Xlib.h>
#include "ewmhview.h"

static Atom net_wm_state;
static Atom net_wm_state_above;
static int atoms_loaded = 0;

static void load_atoms( Display *dpy )
{
    static char *atom_names[] = {
        "_NET_WM_STATE",
        "_NET_WM_STATE_ABOVE"
    };
    Atom atoms_return[ 2 ];

    if( atoms_loaded ) return;
    XInternAtoms( dpy, atom_names, 2, False, atoms_return );
    net_wm_state = atoms_return[ 0 ];
    net_wm_state_above = atoms_return[ 1 ];
    atoms_loaded = 1;
}


static void on_destroy( GtkWidget *widget, gpointer data )
{
    gtk_main_quit();
}

enum
{
    ID_COLUMN,
    NAME_COLUMN,
    X_COLUMN,
    Y_COLUMN,
    W_COLUMN,
    H_COLUMN,
    EVENTS_COLUMN,
    OVERRIDE_COLUMN,
    N_COLUMNS
};

void print_event_mask( char *data, long mask )
{
    int seen = 0;
    *data = 0;
    if( mask & KeyPressMask || mask & KeyReleaseMask ||
        mask & KeymapStateMask) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Key" );
    }
    if( mask & ButtonPressMask ||  mask & ButtonReleaseMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Button" );
    }
    if( mask & EnterWindowMask || mask & LeaveWindowMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Enter" );
    }
    if( mask & PointerMotionMask || mask & PointerMotionHintMask ||
        mask & Button1MotionMask || mask & Button2MotionMask ||
        mask & Button3MotionMask || mask & Button4MotionMask ||
        mask & Button5MotionMask || mask & ButtonMotionMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Mouse" );
    }
    if( mask & ExposureMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Expose" );
    }
    if( mask & VisibilityChangeMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Visibility" );
    }
    if( mask & StructureNotifyMask || mask & ResizeRedirectMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Resize" );
    }
    if( mask & SubstructureNotifyMask || mask & SubstructureRedirectMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Substruct" );
    }
    if( mask & FocusChangeMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Focus" );
    }
    if( mask & ColormapChangeMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Colormap" );
    }
    if( mask & OwnerGrabButtonMask ) {
        if( seen ) { data += sprintf( data, ", " ); } else { seen = 1; }
        data += sprintf( data, "Grab" );
    }
}

void print_name( char *name, Display *dpy, Window win, Window root )
{
    char *wname;
    if( win == root ) {
        sprintf( name, "Root" );
    } else if( !XFetchName( dpy, win, &wname ) ) {
        sprintf( name, " " );
    } else {
        sprintf( name, "%s", wname );
        XFree( wname );
    }
}

GtkTreeStore *store;
GtkWidget *tree;
ewmhview_t *ev;

static gboolean check_events( gpointer data )
{
    Display *dpy = gdk_x11_get_default_xdisplay();
    Window real_root = gdk_x11_get_default_root_xwindow();
    Window win, root, child;
    int rx, ry;
    unsigned int mask;
    GtkTreeIter iter;

    ewmhview_clear( ev );
    gtk_tree_store_clear( store );
    gtk_tree_store_append( store, &iter, 0 );

    win = real_root;

    while( win ) {
        XWindowAttributes attrs;
        int x, y;
        char id[ 30 ];
        char name[ 500 ];
        char xtext[ 30 ];
        char ytext[ 30 ];
        char wtext[ 30 ];
        char htext[ 30 ];
        char override[ 30 ];
        char eventmask[ 255 ];

        XGetWindowAttributes( dpy, win, &attrs );
        sprintf( id, "%x", (unsigned int) win );
        sprintf( xtext, "%d", attrs.x );
        sprintf( ytext, "%d", attrs.y );
        sprintf( wtext, "%d", attrs.width );
        sprintf( htext, "%d", attrs.height );
        print_event_mask( eventmask, attrs.all_event_masks );
        print_name( name, dpy, win, real_root );
        sprintf( override, "%s", attrs.override_redirect ? "Override" : "" );
        ewmhview_load( ev, dpy, win, real_root );

        gtk_tree_store_set( store, &iter,
                            ID_COLUMN, id,
                            NAME_COLUMN, name,
                            X_COLUMN, xtext,
                            Y_COLUMN, ytext,
                            W_COLUMN, wtext,
                            H_COLUMN, htext,
                            EVENTS_COLUMN, eventmask,
                            OVERRIDE_COLUMN, override, -1 );
        gtk_tree_store_append( store, &iter, 0 );
        XQueryPointer(dpy, win, &root, &child, &rx, &ry, &x, &y, &mask );
        win = child;
    }

    return TRUE;
}

void set_state_above( GdkWindow *gdkwin )
{
    Display *dpy = gdk_x11_get_default_xdisplay();
    Window wm_window = gdk_x11_drawable_get_xid( gdkwin );
    XEvent ev;

    load_atoms( dpy );

    ev.type = ClientMessage;
    ev.xclient.window = wm_window;
    ev.xclient.message_type = net_wm_state;
    ev.xclient.format = 32;
    ev.xclient.data.l[ 0 ] = 1;
    ev.xclient.data.l[ 1 ] = net_wm_state_above;
    ev.xclient.data.l[ 2 ] = 0;

    XSendEvent( dpy, DefaultRootWindow( dpy ), False,
                SubstructureNotifyMask|SubstructureRedirectMask, &ev );
}


int main( int argc, char **argv )
{
    GtkWidget *window;
    GtkWidget *notebook;
    GtkWidget *scrolled;
    GtkCellRenderer *renderer;
    GtkTreeViewColumn *column;

    gtk_init( &argc, &argv );

    window = gtk_window_new( GTK_WINDOW_TOPLEVEL );
    gtk_window_set_title( GTK_WINDOW( window ), "X Window Information" );

    g_signal_connect( G_OBJECT( window ), "destroy",
                      G_CALLBACK( on_destroy ), 0 );

    notebook = gtk_notebook_new();
    gtk_container_add( GTK_CONTAINER( window ), notebook );


    scrolled = gtk_scrolled_window_new( 0, 0 );

    store = gtk_tree_store_new( N_COLUMNS,
                                G_TYPE_STRING,  // id
                                G_TYPE_STRING,  // name
                                G_TYPE_STRING,  // x
                                G_TYPE_STRING,  // y
                                G_TYPE_STRING,  // w
                                G_TYPE_STRING,  // h
                                G_TYPE_STRING,  // events
                                G_TYPE_STRING); // override
    tree = gtk_tree_view_new_with_model( GTK_TREE_MODEL( store ) );

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes( "Window", renderer, "text", ID_COLUMN, 0 );
    gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), column );

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes( "Name", renderer, "text", NAME_COLUMN, 0 );
    gtk_tree_view_column_set_min_width( column, 50 );
    gtk_tree_view_column_set_max_width( column, 50 );
    gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), column );

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes( "Override", renderer, "text", OVERRIDE_COLUMN, 0 );
    gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), column );

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes( "X", renderer, "text", X_COLUMN, 0 );
    gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), column );

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes( "Y", renderer, "text", Y_COLUMN, 0 );
    gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), column );
    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes( "Width", renderer, "text", W_COLUMN, 0 );
    gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), column );

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes( "Height", renderer, "text", H_COLUMN, 0 );
    gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), column );

    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes( "Events", renderer, "text", EVENTS_COLUMN, 0 );
    gtk_tree_view_column_set_min_width( column, 100 );
    gtk_tree_view_column_set_max_width( column, 100 );
    gtk_tree_view_column_set_resizable( column, TRUE );
    gtk_tree_view_append_column( GTK_TREE_VIEW( tree ), column );

    gtk_container_add( GTK_CONTAINER( scrolled ), tree );

    gtk_notebook_append_page( GTK_NOTEBOOK( notebook ), scrolled,
                              gtk_label_new( "Window List" ) );


    ev = ewmhview_new();
    gtk_notebook_append_page( GTK_NOTEBOOK( notebook ),
                              ewmhview_get_widget( ev ),
                              gtk_label_new( "Window Manager Hints" ) );

    g_timeout_add( 50, check_events, window );

    gtk_widget_show_all( window );
    set_state_above( ((GtkWidget *) window)->window );

    gtk_main();
    return 0;
}

